/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package mcts;

import java.util.ArrayList;
import java.util.Random;
import tictactoecodingame.Coup;
import tictactoecodingame.Joueur;
import tictactoecodingame.Plateau;

/**
 *
 * @author Lloyd
 */
public class MCTS {
    
    Plateau plateau;
    Noeud racine;
    double coefficientUCT;
    Joueur joueur;
    Joueur adversaire;
    
    //CONSTRUCTEUR
    public MCTS(Plateau plateau, Joueur joueur, Joueur adversaire) {
        this.plateau = plateau;
        this.joueur = joueur;
        this.adversaire = adversaire;
        
        System.out.println(joueur + " / " + adversaire);
        
        plateau.sauvegardePosition(0);
        coefficientUCT = Math.sqrt(2);
        
        Etat etat = new Etat();
        etat.setJoueur(joueur);
        racine = new Noeud();
        racine.setEtat(etat);
    }
    
    //Obtenir le meilleurCoup grâce au MCTS
    public Coup meilleurCoup() {
        double endTime = System.currentTimeMillis() + 20; //L'algorithme tourne pendant 10s
        Noeud noeudPrometteur;
        Noeud nouveauNoeud;
        Joueur gagnantSimulation;
        ArrayList<Coup> coupsPossibles;
        
        while (System.currentTimeMillis() < endTime) {
            
            noeudPrometteur = selection();
            coupsPossibles = plateau.getListeCoups(joueur);
            
            if (coupsPossibles.isEmpty()) {
                propagationArriere(noeudPrometteur, plateau.vainqueur());
            }
            
            else {
                plateau.sauvegardePosition(noeudPrometteur.getEtat().getProfondeur());
                while (!coupsPossibles.isEmpty()) {
                    nouveauNoeud = expansion(noeudPrometteur, coupsPossibles);
                    gagnantSimulation = simulation(nouveauNoeud);
                    //System.out.print(gagnantSimulation + " ");
                    propagationArriere(nouveauNoeud, gagnantSimulation);
                    plateau.restaurePosition(noeudPrometteur.getEtat().getProfondeur());
                }
            }
            
            
            //System.out.println();
            //System.out.println("_");
            
            /*
            if (!racine.getListeEnfant().isEmpty()) {
                for (Noeud enfant : racine.getListeEnfant()) {
                    System.out.print(enfant.getEtat().getNbVictoire() + "/" + enfant.getEtat().getNbVisite() + " | ");
                }
            }
            System.out.println();
            System.out.println(racine.getEtat().getNbVictoire() + "/" + racine.getEtat().getNbVisite());
            */
            
            
        }
        
        plateau.restaurePosition(0);
        return Util.getMeilleurCoup(racine);
    }
    
    //4 PHASE DU MCTS
    
    //Selectionne le noeud "prometteur" grâce à UCT.
    public Noeud selection() {
        plateau.restaurePosition(0);
        Noeud noeudPrometteur = racine;
        while (!noeudPrometteur.getListeEnfant().isEmpty()) {
            noeudPrometteur = UCT.trouverNoeudPrometteur(noeudPrometteur, coefficientUCT);
            plateau.joueCoup(noeudPrometteur.getEtat().getDernierCoup());
            //plateau.sauvegardePosition(noeudPrometteur.getEtat().getProfondeur());
        }
        return noeudPrometteur;
    }
    
    //Une fois le meilleur noeud trouvé, on choisit un noeud au hasard parmi tous les noeuds possible.
    public Noeud expansion(Noeud noeud, ArrayList<Coup> coupsPossibles) {
        Joueur joueurTemp = Util.swapJoueur(noeud.getEtat().getJoueur(), joueur, adversaire);
        int profondeur = noeud.getEtat().getProfondeur();
        Random r = new Random();
        int randomIndex = r.nextInt(coupsPossibles.size());
        Coup randomCoup = coupsPossibles.remove(randomIndex);
        plateau.joueCoup(randomCoup);
        
        Etat etat = new Etat();
        etat.setJoueur(joueurTemp);
        etat.setDernierCoup(randomCoup);
        etat.setProfondeur(profondeur + 1);

        Noeud nouveauNoeud = new Noeud(etat, noeud);
        noeud.getListeEnfant().add(nouveauNoeud);

        return nouveauNoeud;
    }
    
    //A partir du nouveau noeud exploré, on simule une partie de coup aléatoire jusqu'à terminer la partie
    public Joueur simulation(Noeud noeud) {
        Joueur joueurTemp = noeud.getEtat().getJoueur();
        
        while (!plateau.partieTerminee()) {
            Random r = new Random();
            int randomIndex = r.nextInt(plateau.getListeCoups(joueurTemp).size());
            Coup randomCoup = plateau.getListeCoups(joueurTemp).get(randomIndex);
            plateau.joueCoup(randomCoup);
            joueurTemp = Util.swapJoueur(joueurTemp, joueur, adversaire);
        }
        joueurTemp = plateau.vainqueur();
        return joueurTemp;
    }
    
    public void propagationArriere(Noeud noeud, Joueur gagnant) {
        Noeud parent = noeud;
        
        while (parent != null) {
            parent.getEtat().updateNbVisite();
            if (gagnant != null && parent.getEtat().getJoueur() != gagnant) {
                parent.getEtat().updateNbVictoire();
            }
            parent = parent.getParent();
        }
    }
}


/*
//Obtenir le meilleurCoup grâce au MCTS
    public Coup meilleurCoup() {
        double endTime = System.currentTimeMillis() + 100; //L'algorithme tourne pendant 10s
        //System.out.println(racine.getEtat().getCoupsPossibles());
        
        while (System.currentTimeMillis() < endTime) {
            
            Noeud noeudPrometteur;
            Noeud nouveauNoeud;
            Joueur gagnantSimulation;
            
            noeudPrometteur = selection();
            
            //System.out.println(noeudPrometteur.getEtat().getCoupsPossibles());
            //System.out.println(racine.getListeEnfant().size());
            
            ArrayList<Coup> coupsPossibles = plateau.getListeCoups(joueur);
            
            if (coupsPossibles != null) {
                while (!coupsPossibles.isEmpty()) {
                    nouveauNoeud = expansion(noeudPrometteur, coupsPossibles);
                    gagnantSimulation = simulation(nouveauNoeud);
                    //System.out.print(resultat + " ");
                    propagationArriere(nouveauNoeud, gagnantSimulation);
                    //System.out.println(noeudPrometteur.getEtat().getCoupsPossibles());
                }
            }
            
            else {
                gagnantSimulation = plateau.vainqueur();
                propagationArriere(noeudPrometteur, gagnantSimulation);
            }
                
            
            
            
            /*
            if (!plateau.getListeCoups(currentJoueur).isEmpty()) {
                Util.swapJoueur(currentJoueur, joueur, adversaire);
                while(noeudPrometteur.getListeEnfant().size() < plateau.getListeCoups(currentJoueur).size()) {
                    nouveauNoeud = expansion(noeudPrometteur);
                    gagnantSimulation = simulation(nouveauNoeud);
                    //System.out.print(resultat + " ");
                    propagationArriere(nouveauNoeud, gagnantSimulation);
                    //System.out.println(noeudPrometteur.getEtat().getCoupsPossibles());
                }
            }
            else {
                gagnantSimulation = simulation(noeudPrometteur);
                propagationArriere(noeudPrometteur, gagnantSimulation);
            }
            */
            //System.out.println();
            //System.out.println("_");
            /*
            
            if (!racine.getListeEnfant().isEmpty()) {
                for (Noeud enfant : racine.getListeEnfant()) {
                    System.out.print(enfant.getEtat().getNbVictoire() + "/" + enfant.getEtat().getNbVisite() + " ");
                }
            }
            System.out.println();
            System.out.println(racine.getEtat().getNbVictoire() + "/" + racine.getEtat().getNbVisite() + " " );
            */ /*
        } 
       
        return Util.getMeilleurCoup(racine);
    }
*/

/*

 //4 PHASE DU MCTS
    
    //Selectionne le noeud "prometteur" grâce à UCT.
    public Noeud selection() {
        plateau.restaurePosition(0);
        currentJoueur = joueur;
        Noeud noeudPrometteur = racine;
        while (!noeudPrometteur.getListeEnfant().isEmpty()) {
            noeudPrometteur = UCT.trouverNoeudPrometteur(noeudPrometteur, coefficientUCT);
            Util.swapJoueur(currentJoueur, joueur, adversaire);
            plateau.joueCoup(noeudPrometteur.getEtat().getDernierCoup());
            plateau.sauvegardePosition(noeudPrometteur.getEtat().getProfondeur());
        }
        return noeudPrometteur;
    }
    
    //Une fois le meilleur noeud trouvé, on choisit un noeud au hasard parmi tous les noeuds possible.
    public Noeud expansion(Noeud noeud, ArrayList<Coup> coupsPossibles) {
        int profondeur = noeud.getEtat().getProfondeur();
        Random r = new Random();
        int randomIndex = r.nextInt(coupsPossibles.size());
        Coup randomCoup = coupsPossibles.remove(randomIndex);

        plateau.joueCoup(randomCoup);
        plateau.sauvegardePosition(profondeur + 1);
        Util.swapJoueur(currentJoueur, joueur, adversaire);

        Etat etat = new Etat();
        //etat.setNumJoueur(currentJoueur.getIdJoueur());
        etat.setDernierCoup(randomCoup);
        etat.setProfondeur(profondeur + 1);

        Noeud nouveauNoeud = new Noeud(etat, noeud);
        noeud.getListeEnfant().add(nouveauNoeud);

        return nouveauNoeud;
    }
    
    //A partir du nouveau noeud exploré, on simule une partie de coup aléatoire jusqu'à terminer la partie
    public Joueur simulation(Noeud noeud) {
        //int resultat = 0;
        
        Joueur joueurTemp = currentJoueur;
        
        while (!plateau.partieTerminee()) {
            Random r = new Random();
            int randomIndex = r.nextInt(plateau.getListeCoups(joueurTemp).size());
            Coup randomCoup = plateau.getListeCoups(joueurTemp).get(randomIndex);
            plateau.joueCoup(randomCoup);
            Util.swapJoueur(joueurTemp, joueur, adversaire);
        }
        
        return plateau.vainqueur();
    }
    
    public void propagationArriere(Noeud noeud, Joueur gagnant) {
        Noeud parent = noeud;
        Joueur joueurTemp = currentJoueur;
        
        while (parent != null) {
            parent.getEtat().updateNbVisite();
            if (gagnant != null && currentJoueur != gagnant) {
                parent.getEtat().updateNbVictoire();
            }
            parent = parent.getParent();
            Util.swapJoueur(currentJoueur, joueur, adversaire);
        }
    }

    
}
*/